r"""Implement the average precision metrics."""

from __future__ import annotations

__all__ = [
    "average_precision",
    "binary_average_precision",
    "find_label_type",
    "multiclass_average_precision",
    "multilabel_average_precision",
]


import numpy as np
from sklearn import metrics

from arkas.metric.utils import (
    check_label_type,
    contains_nan,
    preprocess_score_binary,
    preprocess_score_multiclass,
    preprocess_score_multilabel,
)


def average_precision(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    label_type: str = "auto",
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the average precision metrics.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        label_type: The type of labels used to evaluate the metrics.
            The valid values are: ``'binary'``, ``'multiclass'``,
            and ``'multilabel'``. If ``'binary'`` or ``'multilabel'``,
            ``y_true`` values  must be ``0`` and ``1``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import average_precision
    >>> # auto
    >>> metrics = average_precision(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_score=np.array([2, -1, 0, 3, 1])
    ... )
    >>> metrics
    {'average_precision': 1.0, 'count': 5}
    >>> # binary
    >>> metrics = average_precision(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_score=np.array([2, -1, 0, 3, 1]),
    ...     label_type="binary",
    ... )
    >>> metrics
    {'average_precision': 1.0, 'count': 5}
    >>> # multiclass
    >>> metrics = average_precision(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_score=np.array(
    ...         [
    ...             [0.7, 0.2, 0.1],
    ...             [0.4, 0.3, 0.3],
    ...             [0.1, 0.8, 0.1],
    ...             [0.2, 0.3, 0.5],
    ...             [0.4, 0.4, 0.2],
    ...             [0.1, 0.2, 0.7],
    ...         ]
    ...     ),
    ...     label_type="multiclass",
    ... )
    >>> metrics
    {'average_precision': array([0.833..., 0.75 , 0.75 ]),
     'count': 6,
     'macro_average_precision': 0.777...,
     'micro_average_precision': 0.75,
     'weighted_average_precision': 0.777...}
    >>> # multilabel
    >>> metrics = average_precision(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_score=np.array([[2, -1, -1], [-1, 1, 2], [0, 2, 3], [3, -2, -4], [1, -3, -5]]),
    ...     label_type="multilabel",
    ... )
    >>> metrics
    {'average_precision': array([1. , 1. , 0.477...]),
     'count': 5,
     'macro_average_precision': 0.825...,
     'micro_average_precision': 0.588...,
     'weighted_average_precision': 0.804...}

    ```
    """
    check_label_type(label_type)
    if label_type == "auto":
        label_type = find_label_type(y_true=y_true, y_score=y_score)
    if label_type == "binary":
        return binary_average_precision(
            y_true=y_true, y_score=y_score, prefix=prefix, suffix=suffix, nan_policy=nan_policy
        )
    if label_type == "multilabel":
        return multilabel_average_precision(
            y_true=y_true, y_score=y_score, prefix=prefix, suffix=suffix, nan_policy=nan_policy
        )
    return multiclass_average_precision(
        y_true=y_true, y_score=y_score, prefix=prefix, suffix=suffix, nan_policy=nan_policy
    )


def binary_average_precision(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the average precision metrics for binary labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, *)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, *)``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import binary_average_precision
    >>> metrics = binary_average_precision(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_score=np.array([2, -1, 0, 3, 1])
    ... )
    >>> metrics
    {'average_precision': 1.0, 'count': 5}

    ```
    """
    y_true, y_score = preprocess_score_binary(
        y_true=y_true, y_score=y_score, drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_score_nan = contains_nan(arr=y_score, nan_policy=nan_policy, name="'y_score'")

    count = y_true.size
    ap = float("nan")
    if count > 0 and not y_true_nan and not y_score_nan:
        ap = float(metrics.average_precision_score(y_true=y_true, y_score=y_score))
    return {f"{prefix}average_precision{suffix}": ap, f"{prefix}count{suffix}": count}


def multiclass_average_precision(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the average precision metrics for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import multiclass_average_precision
    >>> metrics = multiclass_average_precision(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_score=np.array(
    ...         [
    ...             [0.7, 0.2, 0.1],
    ...             [0.4, 0.3, 0.3],
    ...             [0.1, 0.8, 0.1],
    ...             [0.2, 0.3, 0.5],
    ...             [0.4, 0.4, 0.2],
    ...             [0.1, 0.2, 0.7],
    ...         ]
    ...     ),
    ... )
    >>> metrics
    {'average_precision': array([0.833..., 0.75 , 0.75 ]),
     'count': 6,
     'macro_average_precision': 0.777...,
     'micro_average_precision': 0.75,
     'weighted_average_precision': 0.777...}

    ```
    """
    y_true, y_score = preprocess_score_multiclass(y_true, y_score, drop_nan=nan_policy == "omit")
    return _average_precision(
        y_true=y_true, y_score=y_score, prefix=prefix, suffix=suffix, nan_policy=nan_policy
    )


def multilabel_average_precision(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the average precision metrics for multilabel labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, n_classes)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import multilabel_average_precision
    >>> metrics = multilabel_average_precision(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_score=np.array([[2, -1, -1], [-1, 1, 2], [0, 2, 3], [3, -2, -4], [1, -3, -5]]),
    ... )
    >>> metrics
    {'average_precision': array([1. , 1. , 0.477...]),
     'count': 5,
     'macro_average_precision': 0.825...,
     'micro_average_precision': 0.588...,
     'weighted_average_precision': 0.804...}

    ```
    """
    y_true, y_score = preprocess_score_multilabel(y_true, y_score, drop_nan=nan_policy == "omit")
    return _average_precision(
        y_true=y_true, y_score=y_score, prefix=prefix, suffix=suffix, nan_policy=nan_policy
    )


def _average_precision(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the average precision metrics for multilabel or multiclass
    labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, n_classes)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.
    """
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_score_nan = contains_nan(arr=y_score, nan_policy=nan_policy, name="'y_score'")

    n_samples = y_true.shape[0]
    macro, micro, weighted = float("nan"), float("nan"), float("nan")
    ap = np.array([])
    if n_samples > 0 and not y_true_nan and not y_score_nan:
        macro = metrics.average_precision_score(y_true=y_true, y_score=y_score, average="macro")
        micro = metrics.average_precision_score(y_true=y_true, y_score=y_score, average="micro")
        weighted = metrics.average_precision_score(
            y_true=y_true, y_score=y_score, average="weighted"
        )
        ap = np.asarray(
            metrics.average_precision_score(y_true=y_true, y_score=y_score, average=None)
        ).ravel()

    return {
        f"{prefix}average_precision{suffix}": ap,
        f"{prefix}count{suffix}": n_samples,
        f"{prefix}macro_average_precision{suffix}": float(macro),
        f"{prefix}micro_average_precision{suffix}": float(micro),
        f"{prefix}weighted_average_precision{suffix}": float(weighted),
    }


def find_label_type(y_true: np.ndarray, y_score: np.ndarray) -> str:
    r"""Try to find the label type automatically based on the arrays'
    shape.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.

    Returns:
        The label type.

    Raises:
        RuntimeError: if the label type cannot be found automatically.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.classification.ap import find_label_type
    >>> # binary
    >>> label_type = find_label_type(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_score=np.array([2, -1, 0, 3, 1]),
    ... )
    >>> label_type
    'binary'

    ```
    """
    if y_true.ndim == 1 and y_score.ndim == 2:
        return "multiclass"
    if y_true.ndim == 2 and y_score.ndim == 2:
        return "multilabel"
    if y_true.ndim == 1 and y_score.ndim == 1:
        return "binary"
    msg = "Could not find the label type"
    raise RuntimeError(msg)



r"""Implement the recall result."""

from __future__ import annotations

__all__ = [
    "BaseRecallResult",
    "BinaryRecallResult",
    "MulticlassRecallResult",
    "MultilabelRecallResult",
    "RecallResult",
]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric import recall
from arkas.metric.classification.recall import (
    binary_recall,
    find_label_type,
    multiclass_recall,
    multilabel_recall,
)
from arkas.metric.figure import binary_precision_recall_curve
from arkas.metric.utils import check_label_type, check_nan_policy, check_same_shape_pred
from arkas.result.base import BaseResult

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    import numpy as np


class RecallResult(BaseResult):
    r"""Implement the recall result.

    This result can be used in 3 different settings:

    - binary: ``y_true`` must be an array of shape ``(n_samples,)``
        with ``0`` and ``1`` values, and ``y_pred`` must be an array
        of shape ``(n_samples,)``.
    - multiclass: ``y_true`` must be an array of shape ``(n_samples,)``
        with values in ``{0, ..., n_classes-1}``, and ``y_pred`` must
        be an array of shape ``(n_samples,)``.
    - multilabel: ``y_true`` must be an array of shape
        ``(n_samples, n_classes)`` with ``0`` and ``1`` values, and
        ``y_pred`` must be an array of shape
        ``(n_samples, n_classes)``.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        y_pred: The predicted labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import RecallResult
    >>> # binary
    >>> result = RecallResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_pred=np.array([1, 0, 0, 1, 1]),
    ...     label_type="binary",
    ... )
    >>> result
    RecallResult(y_true=(5,), y_pred=(5,), label_type='binary', nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'recall': 1.0}
    >>> # multilabel
    >>> result = RecallResult(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 0], [0, 1, 1], [0, 1, 1], [1, 0, 0], [1, 0, 0]]),
    ...     label_type="multilabel",
    ... )
    >>> result
    RecallResult(y_true=(5, 3), y_pred=(5, 3), label_type='multilabel', nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5,
     'macro_recall': 0.666...,
     'micro_recall': 0.625,
     'recall': array([1., 1., 0.]),
     'weighted_recall': 0.625}
    >>> # multiclass
    >>> result = RecallResult(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_pred=np.array([0, 0, 1, 1, 2, 2]),
    ...     label_type="multiclass",
    ... )
    >>> result
    RecallResult(y_true=(6,), y_pred=(6,), label_type='multiclass', nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 6,
     'macro_recall': 1.0,
     'micro_recall': 1.0,
     'recall': array([1., 1., 1.]),
     'weighted_recall': 1.0}
    >>> # auto
    >>> result = RecallResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    RecallResult(y_true=(5,), y_pred=(5,), label_type='binary', nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'recall': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        label_type: str = "auto",
        nan_policy: str = "propagate",
    ) -> None:
        self._y_true = y_true
        self._y_pred = y_pred
        self._label_type = (
            find_label_type(y_true=y_true, y_pred=y_pred) if label_type == "auto" else label_type
        )
        self._check_inputs()

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_pred": self._y_pred.shape,
                "label_type": self._label_type,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def label_type(self) -> str:
        return self._label_type

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_pred(self) -> np.ndarray:
        return self._y_pred

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return recall(
            y_true=self._y_true,
            y_pred=self._y_pred,
            label_type=self._label_type,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_pred, other.y_pred, equal_nan=equal_nan)
            and self.label_type == other.label_type
            and self.nan_policy == other.nan_policy
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}

    def _check_inputs(self) -> None:
        if self._y_true.ndim not in {1, 2}:
            msg = (
                f"'y_true' must be a 1d or 2d array but received an array of shape: "
                f"{self._y_true.shape}"
            )
            raise ValueError(msg)
        if self._y_pred.ndim not in {1, 2}:
            msg = (
                f"'y_pred' must be a 1d or 2d array but received an array of shape: "
                f"{self._y_pred.shape}"
            )
            raise ValueError(msg)
        check_label_type(self._label_type)


class BaseRecallResult(BaseResult):
    r"""Implement the base class to implement the recall results.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryRecallResult
    >>> result = BinaryRecallResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    BinaryRecallResult(y_true=(5,), y_pred=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'recall': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        self._y_true = y_true
        self._y_pred = y_pred

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_pred": self._y_pred.shape,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_pred(self) -> np.ndarray:
        return self._y_pred

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_pred, other.y_pred, equal_nan=equal_nan)
            and self.nan_policy == other.nan_policy
        )


class BinaryRecallResult(BaseRecallResult):
    r"""Implement the recall result for binary labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, *)`` with ``0`` and
            ``1`` values.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples, *)`` with ``0`` and ``1`` values.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryRecallResult
    >>> result = BinaryRecallResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    BinaryRecallResult(y_true=(5,), y_pred=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'recall': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_pred(y_true, y_pred)
        super().__init__(y_true=y_true.ravel(), y_pred=y_pred.ravel(), nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return binary_recall(
            y_true=self._y_true,
            y_pred=self._y_pred,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(self, prefix: str = "", suffix: str = "") -> dict[str, plt.Figure]:
        fig = binary_precision_recall_curve(
            y_true=self._y_true,
            y_pred=self._y_pred,
        )
        if fig is None:
            return {}
        return {f"{prefix}precision_recall{suffix}": fig}


class MulticlassRecallResult(BaseRecallResult):
    r"""Implement the recall result for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, *)`` with values in
            ``{0, ..., n_classes-1}``.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples, *)`` with values in
            ``{0, ..., n_classes-1}``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MulticlassRecallResult
    >>> result = MulticlassRecallResult(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_pred=np.array([0, 0, 1, 1, 2, 2]),
    ... )
    >>> result
    MulticlassRecallResult(y_true=(6,), y_pred=(6,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 6,
     'macro_recall': 1.0,
     'micro_recall': 1.0,
     'recall': array([1., 1., 1.]),
     'weighted_recall': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_pred(y_true, y_pred)
        super().__init__(y_true=y_true.ravel(), y_pred=y_pred.ravel(), nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return multiclass_recall(
            y_true=self._y_true,
            y_pred=self._y_pred,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}


class MultilabelRecallResult(BaseRecallResult):
    r"""Implement the recall result for multilabel labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, n_classes)`` with ``0``
            and ``1`` values.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples, n_classes)`` with ``0`` and ``1``
            values.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MultilabelRecallResult
    >>> result = MultilabelRecallResult(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ... )
    >>> result
    MultilabelRecallResult(y_true=(5, 3), y_pred=(5, 3), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5,
     'macro_recall': 1.0,
     'micro_recall': 1.0,
     'recall': array([1., 1., 1.]),
     'weighted_recall': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_pred(y_true, y_pred)
        super().__init__(y_true=y_true, y_pred=y_pred, nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return multilabel_recall(
            y_true=self._y_true,
            y_pred=self._y_pred,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}



r"""Contain the implementation of a simple reporter that evaluates
results and writes them in a HTML file."""

from __future__ import annotations

__all__ = ["EvalReporter"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from coola.utils.path import sanitize_path
from grizz.ingestor import BaseIngestor, setup_ingestor
from grizz.transformer import BaseTransformer, setup_transformer
from iden.io import save_text

from arkas.evaluator.base import BaseEvaluator, setup_evaluator
from arkas.reporter.base import BaseReporter
from arkas.reporter.utils import create_html_report
from arkas.section import ResultSection

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


class EvalReporter(BaseReporter):
    r"""Implement a simple reporter that evaluates results and writes
    them in a HTML file.

    Args:
        ingestor: The ingestor or its configuration.
        transformer: The data transformer or its configuration.
        evaluator: The evaluator or its configuration.
        report_path: The path where to save the HTML report.
        max_toc_depth: The maximum level to show in the
            table of content.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> from pathlib import Path
    >>> import polars as pl
    >>> from arkas.evaluator import AccuracyEvaluator
    >>> from grizz.ingestor import Ingestor
    >>> from grizz.transformer import SequentialTransformer
    >>> from arkas.reporter import EvalReporter
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     reporter = EvalReporter(
    ...         ingestor=Ingestor(
    ...             pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
    ...         ),
    ...         transformer=SequentialTransformer(transformers=[]),
    ...         evaluator=AccuracyEvaluator(y_true="target", y_pred="pred"),
    ...         report_path=Path(tmpdir).joinpath("report.html"),
    ...     )
    ...     reporter.generate()
    ...

    ```
    """

    def __init__(
        self,
        ingestor: BaseIngestor | dict,
        transformer: BaseTransformer | dict,
        evaluator: BaseEvaluator | dict,
        report_path: Path | str,
        max_toc_depth: int = 6,
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        logger.info(f"ingestor:\n{ingestor}")
        self._transformer = setup_transformer(transformer)
        logger.info(f"transformer:\n{transformer}")
        self._evaluator = setup_evaluator(evaluator)
        logger.info(f"evaluator:\n{evaluator}")
        self._report_path = sanitize_path(report_path)
        self._max_toc_depth = int(max_toc_depth)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "ingestor": self._ingestor,
                    "transformer": self._transformer,
                    "evaluator": self._evaluator,
                    "report_path": self._report_path,
                    "max_toc_depth": self._max_toc_depth,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self) -> None:
        logger.info("Ingesting the DataFrame...")
        frame = self._ingestor.ingest()
        logger.info(f"Transforming the DataFrame {frame.shape}...")
        frame = self._transformer.transform(frame)
        logger.info(f"Analyzing the DataFrame {frame.shape}...")
        result = self._evaluator.evaluate(frame)
        logger.info("Creating the HTML report with the results...")
        section = ResultSection(result)
        report = create_html_report(
            toc=section.generate_html_body(),
            body=section.generate_html_toc(max_depth=self._max_toc_depth),
        )
        logger.info(f"Saving the HTML report at {self._report_path}...")
        save_text(report, self._report_path, exist_ok=True)



r"""Contain results."""

from __future__ import annotations

__all__ = [
    "AccuracyResult",
    "AveragePrecisionResult",
    "BalancedAccuracyResult",
    "BaseResult",
    "BinaryAveragePrecisionResult",
    "BinaryClassificationResult",
    "BinaryConfusionMatrixResult",
    "BinaryFbetaScoreResult",
    "BinaryJaccardResult",
    "BinaryPrecisionResult",
    "BinaryRecallResult",
    "BinaryRocAucResult",
    "EmptyResult",
    "EnergyDistanceResult",
    "JensenShannonDivergenceResult",
    "KLDivResult",
    "MappingResult",
    "MeanAbsoluteErrorResult",
    "MeanAbsolutePercentageErrorResult",
    "MeanSquaredErrorResult",
    "MeanSquaredLogErrorResult",
    "MeanTweedieDevianceResult",
    "MedianAbsoluteErrorResult",
    "MulticlassAveragePrecisionResult",
    "MulticlassConfusionMatrixResult",
    "MulticlassFbetaScoreResult",
    "MulticlassJaccardResult",
    "MulticlassPrecisionResult",
    "MulticlassRecallResult",
    "MulticlassRocAucResult",
    "MultilabelAveragePrecisionResult",
    "MultilabelConfusionMatrixResult",
    "MultilabelFbetaScoreResult",
    "MultilabelJaccardResult",
    "MultilabelPrecisionResult",
    "MultilabelRecallResult",
    "MultilabelRocAucResult",
    "PearsonCorrelationResult",
    "PrecisionResult",
    "R2ScoreResult",
    "RecallResult",
    "RegressionErrorResult",
    "Result",
    "RootMeanSquaredErrorResult",
    "SequentialResult",
    "SpearmanCorrelationResult",
    "WassersteinDistanceResult",
]

from arkas.result.accuracy import AccuracyResult
from arkas.result.ap import (
    AveragePrecisionResult,
    BinaryAveragePrecisionResult,
    MulticlassAveragePrecisionResult,
    MultilabelAveragePrecisionResult,
)
from arkas.result.balanced_accuracy import BalancedAccuracyResult
from arkas.result.base import BaseResult
from arkas.result.binary_classification import BinaryClassificationResult
from arkas.result.confmat import (
    BinaryConfusionMatrixResult,
    MulticlassConfusionMatrixResult,
    MultilabelConfusionMatrixResult,
)
from arkas.result.energy import EnergyDistanceResult
from arkas.result.fbeta import (
    BinaryFbetaScoreResult,
    MulticlassFbetaScoreResult,
    MultilabelFbetaScoreResult,
)
from arkas.result.jaccard import (
    BinaryJaccardResult,
    MulticlassJaccardResult,
    MultilabelJaccardResult,
)
from arkas.result.jensen_shannon import JensenShannonDivergenceResult
from arkas.result.kl import KLDivResult
from arkas.result.mae import MeanAbsoluteErrorResult
from arkas.result.mape import MeanAbsolutePercentageErrorResult
from arkas.result.mapping import MappingResult
from arkas.result.median_error import MedianAbsoluteErrorResult
from arkas.result.mse import MeanSquaredErrorResult
from arkas.result.msle import MeanSquaredLogErrorResult
from arkas.result.pearson import PearsonCorrelationResult
from arkas.result.precision import (
    BinaryPrecisionResult,
    MulticlassPrecisionResult,
    MultilabelPrecisionResult,
    PrecisionResult,
)
from arkas.result.r2 import R2ScoreResult
from arkas.result.recall import (
    BinaryRecallResult,
    MulticlassRecallResult,
    MultilabelRecallResult,
    RecallResult,
)
from arkas.result.regression import RegressionErrorResult
from arkas.result.rmse import RootMeanSquaredErrorResult
from arkas.result.roc_auc import (
    BinaryRocAucResult,
    MulticlassRocAucResult,
    MultilabelRocAucResult,
)
from arkas.result.sequential import SequentialResult
from arkas.result.spearman import SpearmanCorrelationResult
from arkas.result.tweedie_deviance import MeanTweedieDevianceResult
from arkas.result.vanilla import EmptyResult, Result
from arkas.result.wasserstein import WassersteinDistanceResult



r"""Contain the implementation for plotly figures."""

from __future__ import annotations

__all__ = ["PlotlyFigure", "PlotlyFigureConfig"]

import sys
from typing import Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.figure.base import BaseFigure, BaseFigureConfig
from arkas.utils.imports import check_plotly, is_plotly_available

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import (
        Self,  # use backport because it was added in python 3.11
    )

if is_plotly_available():
    from plotly.graph_objects import Figure
else:  # pragma: no cover
    from unittest.mock import Mock

    Figure = Mock()


class PlotlyFigure(BaseFigure):
    r"""Implement the plotly figure.

    Args:
        figure: The plotly figure.
        reactive: If ``True``, the generated is configured to be
            reactive to the screen size.

    Example usage:

    ```pycon

    >>> from arkas.figure import PlotlyFigure
    >>> fig = PlotlyFigure(None)
    >>> fig
    PlotlyFigure(reactive=True)

    ```
    """

    def __init__(self, figure: Figure, reactive: bool = True) -> None:
        check_plotly()
        self._figure = figure
        self._reactive = reactive

    def __repr__(self) -> str:
        args = repr_mapping_line({"reactive": self._reactive})
        return f"{self.__class__.__qualname__}({args})"

    @property
    def figure(self) -> Figure:
        return self._figure

    def close(self) -> None:
        pass

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.figure, other.figure, equal_nan=equal_nan)
            and self._reactive == other._reactive
        )

    def set_reactive(self, reactive: bool) -> Self:
        return self.__class__(figure=self.figure, reactive=reactive)

    def to_html(self) -> str:
        data = ""
        style = 'style="width:100%; height:auto;" ' if self._reactive else ""
        return f'<img {style}src="data:image/png;charset=utf-8;base64, {data}">'


class PlotlyFigureConfig(BaseFigureConfig):
    r"""Implement the plotly figure config.

    Args:
        **kwargs: Additional keyword arguments to pass to plotly
            functions. The valid arguments depend on the context.

    Example usage:

    ```pycon

    >>> from arkas.figure import PlotlyFigureConfig
    >>> config = PlotlyFigureConfig()
    >>> config
    PlotlyFigureConfig()

    ```
    """

    def __init__(self, **kwargs: Any) -> None:
        check_plotly()
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(self.get_args())
        return f"{self.__class__.__qualname__}({args})"

    @classmethod
    def backend(cls) -> str:
        return "plotly"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self.get_args(), other.get_args(), equal_nan=equal_nan)

    def get_args(self) -> dict:
        return self._kwargs



r"""Contain the implementation of a section that analyze precision
results."""

from __future__ import annotations

__all__ = ["BinaryPrecisionSection", "create_section_template"]

import logging
from typing import TYPE_CHECKING, Any

from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping
from jinja2 import Template

from arkas.section.base import BaseSection
from arkas.utils.html import GO_TO_TOP, render_toc, tags2id, tags2title, valid_h_tag

if TYPE_CHECKING:
    from collections.abc import Sequence

    from arkas.result import BaseResult


logger = logging.getLogger(__name__)


class BinaryPrecisionSection(BaseSection):
    r"""Implement a section that analyze precision results.

    Args:
        result: The data structure containing the results.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.section import BinaryPrecisionSection
    >>> from arkas.result import BinaryPrecisionResult
    >>> section = BinaryPrecisionSection(
    ...     result=BinaryPrecisionResult(
    ...         y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ...     )
    ... )
    >>> section
    BinaryPrecisionSection(
      (result): BinaryPrecisionResult(y_true=(5,), y_pred=(5,), nan_policy='propagate')
    )
    >>> section.generate_html_body()

    ```
    """

    def __init__(self, result: BaseResult) -> None:
        self._result = result

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping({"result": self._result}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def __str__(self) -> str:
        args = str_indent(str_mapping({"result": self._result}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._result.equal(other._result)

    def generate_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info("Generating the binary precision section...")
        metrics = self._result.compute_metrics()
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "precision": f"{metrics.get('precision', float('nan')):.4f}",
                "count": f"{metrics.get('count', 0):,}",
            }
        )

    def generate_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from arkas.section.accuracy import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">

<ul>
  <li>precision: {{precision}}</li>
  <li>count: {{count}}</li>
</ul>

<p style="margin-top: 1rem;">
"""



r"""Implement the Spearman correlation metrics."""

from __future__ import annotations

__all__ = ["spearmanr"]


from typing import TYPE_CHECKING

from arkas.metric.utils import contains_nan, preprocess_same_shape_arrays
from arkas.utils.imports import check_scipy, is_scipy_available

if is_scipy_available():
    from scipy import stats

if TYPE_CHECKING:
    import numpy as np


def spearmanr(
    x: np.ndarray,
    y: np.ndarray,
    *,
    alternative: str = "two-sided",
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the Spearman correlation coefficient and p-value for
    testing non-correlation.

    Args:
        x: The first input array.
        y: The second input array.
        alternative: The alternative hypothesis. Default is 'two-sided'.
            The following options are available:
            - 'two-sided': the correlation is nonzero
            - 'less': the correlation is negative (less than zero)
            - 'greater': the correlation is positive (greater than zero)
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import spearmanr
    >>> spearmanr(
    ...     x=np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]),
    ...     y=np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]),
    ... )
    {'count': 9, 'spearman_coeff': 1.0, 'spearman_pvalue': 0.0}

    ```
    """
    check_scipy()
    x, y = preprocess_same_shape_arrays(
        arrays=[x.ravel(), y.ravel()], drop_nan=nan_policy == "omit"
    )
    x_nan = contains_nan(arr=x, nan_policy=nan_policy, name="'x'")
    y_nan = contains_nan(arr=y, nan_policy=nan_policy, name="'y'")

    count = x.size
    coeff, pvalue = float("nan"), float("nan")
    if count > 0 and not x_nan and not y_nan:
        result = stats.spearmanr(x, y, alternative=alternative)
        coeff, pvalue = float(result.statistic), float(result.pvalue)
    return {
        f"{prefix}count{suffix}": count,
        f"{prefix}spearman_coeff{suffix}": coeff,
        f"{prefix}spearman_pvalue{suffix}": pvalue,
    }



r"""Implement the accuracy result."""

from __future__ import annotations

__all__ = ["AccuracyResult"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric.classification.accuracy import accuracy
from arkas.metric.utils import check_nan_policy, check_same_shape_pred
from arkas.result.base import BaseResult

if TYPE_CHECKING:
    import numpy as np


class AccuracyResult(BaseResult):
    r"""Implement the accuracy result.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` where the values
            are in ``{0, ..., n_classes-1}``.
        y_pred: The predicted labels. This input must be an
            array of shape ``(n_samples,)`` where the values are
            in ``{0, ..., n_classes-1}``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import AccuracyResult
    >>> result = AccuracyResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    AccuracyResult(y_true=(5,), y_pred=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'accuracy': 1.0, 'count_correct': 5, 'count_incorrect': 0, 'count': 5, 'error': 0.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        self._y_true = y_true.ravel()
        self._y_pred = y_pred.ravel()
        check_same_shape_pred(y_true=self._y_true, y_pred=self._y_pred)

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_pred": self._y_pred.shape,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_pred(self) -> np.ndarray:
        return self._y_pred

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return accuracy(
            y_true=self._y_true,
            y_pred=self._y_pred,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_pred, other.y_pred, equal_nan=equal_nan)
            and self.nan_policy == other.nan_policy
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}



r"""Contain the default figure config."""

from __future__ import annotations

__all__ = ["DefaultFigureConfig"]

from typing import Any

from arkas.figure.base import BaseFigureConfig


class DefaultFigureConfig(BaseFigureConfig):
    r"""Implement the default figure config.

    Example usage:

    ```pycon

    >>> from arkas.figure import DefaultFigureConfig
    >>> config = DefaultFigureConfig()
    >>> config
    DefaultFigureConfig()

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    @classmethod
    def backend(cls) -> str:
        return "default"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:  # noqa: ARG002
        return isinstance(other, self.__class__)

    def get_args(self) -> dict:
        return {}



r"""Contain an analyzer that generates an output with the given custom
content."""

from __future__ import annotations

__all__ = ["ContentAnalyzer"]

import logging
from typing import TYPE_CHECKING

from arkas.analyzer.lazy import BaseLazyAnalyzer
from arkas.output.content import ContentOutput

if TYPE_CHECKING:
    import polars as pl


logger = logging.getLogger(__name__)


class ContentAnalyzer(BaseLazyAnalyzer):
    r"""Implement an analyzer that generates an output with the given
    custom content.

    Args:
        content: The content to use in the HTML code.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import ContentAnalyzer
    >>> analyzer = ContentAnalyzer(content="meow")
    >>> analyzer
    ContentAnalyzer()
    >>> frame = pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
    >>> output = analyzer.analyze(frame)
    >>> output
    ContentOutput(
      (content): ContentGenerator()
      (evaluator): Evaluator(count=0)
      (plotter): Plotter(count=0)
    )

    ```
    """

    def __init__(self, content: str) -> None:
        self._content = content

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _analyze(self, frame: pl.DataFrame) -> ContentOutput:  # noqa: ARG002
        return ContentOutput(self._content)



r"""Contain the base class to implement a section."""

from __future__ import annotations

__all__ = ["BaseResult"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualNanHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

if TYPE_CHECKING:
    from coola.equality import EqualityConfig
    from matplotlib import pyplot as plt


class BaseResult(ABC):
    r"""Define the base class to manage results.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import AccuracyResult
    >>> result = AccuracyResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    AccuracyResult(y_true=(5,), y_pred=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'accuracy': 1.0, 'count_correct': 5, 'count_incorrect': 0, 'count': 5, 'error': 0.0}

    ```
    """

    @abstractmethod
    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict:
        r"""Return the metrics associated to the result.

        Args:
            prefix: The key prefix in the returned dictionary.
            suffix: The key suffix in the returned dictionary.

        Returns:
            The metrics.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.result import AccuracyResult
        >>> result = AccuracyResult(
        ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
        ... )
        >>> result.compute_metrics()
        {'accuracy': 1.0, 'count_correct': 5, 'count_incorrect': 0, 'count': 5, 'error': 0.0}

        ```
        """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two results are equal or not.

        Args:
            other: The other result to compare.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two results are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.result import AccuracyResult
        >>> res1 = AccuracyResult(
        ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
        ... )
        >>> res2 = AccuracyResult(
        ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
        ... )
        >>> res3 = AccuracyResult(
        ...     y_true=np.array([1, 0, 0, 0, 0]), y_pred=np.array([1, 0, 0, 1, 1])
        ... )
        >>> res1.equal(res2)
        True
        >>> res1.equal(res3)
        False

        ```
        """

    @abstractmethod
    def generate_figures(self, prefix: str = "", suffix: str = "") -> dict[str, plt.Figure]:
        r"""Return the figures associated to the result.

        Args:
            prefix: The key prefix in the returned dictionary.
            suffix: The key suffix in the returned dictionary.

        Returns:
            The figures.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from arkas.result import AccuracyResult
        >>> result = AccuracyResult(
        ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
        ... )
        >>> result.generate_figures()
        {}

        ```
        """


class ResultEqualityComparator(BaseEqualityComparator[BaseResult]):
    r"""Implement an equality comparator for ``BaseResult`` objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualNanHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> ResultEqualityComparator:
        return self.__class__()

    def equal(self, actual: BaseResult, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BaseResult):  # pragma: no cover
    EqualityTester.add_comparator(BaseResult, ResultEqualityComparator())



r"""Contain figures to show precision-recall curves."""

from __future__ import annotations

__all__ = ["binary_precision_recall_curve"]

from typing import TYPE_CHECKING, Any

import matplotlib.pyplot as plt

from arkas import plot
from arkas.metric.utils import check_same_shape_pred

if TYPE_CHECKING:
    import numpy as np


def binary_precision_recall_curve(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    name: str = "model",
    plot_chance_level: bool = True,
    **kwargs: Any,
) -> plt.Figure | None:
    r"""Return a figure with the precision-recall curve for binary
    labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` with ``0`` and
            ``1`` values.
        y_pred: The predicted labels. This input must be an array of
            shape ``(n_samples,)`` with ``0`` and ``1`` values.
        name: The name for labeling curve.
        plot_chance_level: Whether to plot the chance level.
            The chance level is the prevalence of the positive label
            computed from the data.
        **kwargs: Arbitrary keyword arguments that are passed to
            ``PrecisionRecallDisplay.from_predictions``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric.figure import binary_precision_recall_curve
    >>> fig = binary_precision_recall_curve(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )

    ```
    """
    check_same_shape_pred(y_true, y_pred)
    if y_true.size == 0:
        return None

    fig, ax = plt.subplots()
    plot.binary_precision_recall_curve(
        ax=ax,
        y_true=y_true,
        y_pred=y_pred,
        name=name,
        plot_chance_level=plot_chance_level,
        **kwargs,
    )
    return fig



r"""Implement the accuracy state."""

from __future__ import annotations

__all__ = ["PrecisionRecallState"]

import sys
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric.classification.precision import find_label_type
from arkas.metric.utils import check_label_type, check_same_shape_pred
from arkas.state.base import BaseState

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import (
        Self,  # use backport because it was added in python 3.11
    )

if TYPE_CHECKING:
    import numpy as np


class PrecisionRecallState(BaseState):
    r"""Implement a state for precision-recall-based metrics.

    This state can be used in 3 different settings:

    - binary: ``y_true`` must be an array of shape ``(n_samples,)``
        with ``0`` and ``1`` values, and ``y_pred`` must be an array
        of shape ``(n_samples,)``.
    - multiclass: ``y_true`` must be an array of shape ``(n_samples,)``
        with values in ``{0, ..., n_classes-1}``, and ``y_pred`` must
        be an array of shape ``(n_samples,)``.
    - multilabel: ``y_true`` must be an array of shape
        ``(n_samples, n_classes)`` with ``0`` and ``1`` values, and
        ``y_pred`` must be an array of shape
        ``(n_samples, n_classes)``.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        y_pred: The predicted labels. This input must
            be an array of shape ``(n_samples,)`` or
            ``(n_samples, n_classes)``.
        y_true_name: The name associated to the ground truth target
            labels.
        y_pred_name: The name associated to the predicted labels.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.state import PrecisionRecallState
    >>> state = PrecisionRecallState(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_pred=np.array([1, 0, 0, 1, 1]),
    ...     y_true_name="target",
    ...     y_pred_name="pred",
    ... )
    >>> state
    PrecisionRecallState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred', label_type='binary')

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        y_true_name: str,
        y_pred_name: str,
        label_type: str = "auto",
    ) -> None:
        self._y_true = y_true
        self._y_pred = y_pred
        self._y_true_name = y_true_name
        self._y_pred_name = y_pred_name
        self._label_type = (
            find_label_type(y_true=y_true, y_pred=y_pred) if label_type == "auto" else label_type
        )
        self._check_args()

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_pred": self._y_pred.shape,
                "y_true_name": self._y_true_name,
                "y_pred_name": self._y_pred_name,
                "label_type": self._label_type,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_pred(self) -> np.ndarray:
        return self._y_pred

    @property
    def y_true_name(self) -> str:
        return self._y_true_name

    @property
    def y_pred_name(self) -> str:
        return self._y_pred_name

    @property
    def label_type(self) -> str:
        return self._label_type

    def clone(self, deep: bool = True) -> Self:
        return self.__class__(
            y_true=self._y_true.copy() if deep else self._y_true,
            y_pred=self._y_pred.copy() if deep else self._y_pred,
            y_true_name=self._y_true_name,
            y_pred_name=self._y_pred_name,
            label_type=self._label_type,
        )

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_pred, other.y_pred, equal_nan=equal_nan)
            and self.y_true_name == other.y_true_name
            and self.y_pred_name == other.y_pred_name
            and self.label_type == other.label_type
        )

    def _check_args(self) -> None:
        if self._y_true.ndim not in {1, 2}:
            msg = (
                f"'y_true' must be a 1d or 2d array but received an array of shape: "
                f"{self._y_true.shape}"
            )
            raise ValueError(msg)
        if self._y_pred.ndim not in {1, 2}:
            msg = (
                f"'y_pred' must be a 1d or 2d array but received an array of shape: "
                f"{self._y_pred.shape}"
            )
            raise ValueError(msg)
        check_same_shape_pred(y_true=self._y_true, y_pred=self._y_pred)
        check_label_type(self._label_type)



r"""Contain the implementation for matplotlib figures."""

from __future__ import annotations

__all__ = ["HtmlFigure"]

import sys
from typing import Any

from arkas.figure.base import BaseFigure

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import (
        Self,  # use backport because it was added in python 3.11
    )


class HtmlFigure(BaseFigure):
    r"""Implement a simple HTML figure.

    Args:
        figure: The HTML code of the figure.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from arkas.figure import HtmlFigure
    >>> fig = HtmlFigure()
    >>> fig
    HtmlFigure()

    ```
    """

    def __init__(self, figure: str = "") -> None:
        self._figure = figure

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def close(self) -> None:
        pass

    def equal(self, other: Any, equal_nan: bool = False) -> bool:  # noqa: ARG002
        if not isinstance(other, self.__class__):
            return False
        return self._figure == other._figure

    def set_reactive(self, reactive: bool) -> Self:  # noqa: ARG002
        return self

    def to_html(self) -> str:
        return self._figure



r"""Define a base class to implement lazy analyzers."""

from __future__ import annotations

__all__ = ["BaseInNLazyAnalyzer", "BaseLazyAnalyzer"]

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line
from grizz.utils.column import (
    check_column_missing_policy,
    check_missing_columns,
    find_common_columns,
    find_missing_columns,
)

from arkas.analyzer.base import BaseAnalyzer

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

    from arkas.output import BaseOutput

logger = logging.getLogger(__name__)


class BaseLazyAnalyzer(BaseAnalyzer):
    r"""Define a base class to implement a lazy analyzer.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import DataFrameSummaryAnalyzer
    >>> analyzer = DataFrameSummaryAnalyzer()
    >>> analyzer
    DataFrameSummaryAnalyzer(top=5, sort=False)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 0, 1],
    ...         "col2": [1, 0, 1, 0],
    ...         "col3": [1, 1, 1, 1],
    ...     },
    ...     schema={"col1": pl.Int64, "col2": pl.Int64, "col3": pl.Int64},
    ... )
    >>> output = analyzer.analyze(frame)
    >>> output
    DataFrameSummaryOutput(shape=(4, 3), top=5)

    ```
    """

    def analyze(self, frame: pl.DataFrame, lazy: bool = True) -> BaseOutput:
        output = self._analyze(frame)
        if not lazy:
            output = output.compute()
        return output

    @abstractmethod
    def _analyze(self, frame: pl.DataFrame) -> BaseOutput:
        r"""Analyze the DataFrame.

        Args:
            frame: The DataFrame to analyze.

        Returns:
            The generated output.
        """


class BaseInNLazyAnalyzer(BaseAnalyzer):
    r"""Define a base class to implement analyzers that analyze
    DataFrames by using multiple input columns.

    Args:
        columns: The columns to analyze. If ``None``, it analyzes all
            the columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.analyzer import ColumnCooccurrenceAnalyzer
    >>> analyzer = ColumnCooccurrenceAnalyzer()
    >>> analyzer
    ColumnCooccurrenceAnalyzer(columns=None, exclude_columns=(), missing_policy='raise', ignore_self=False, figure_config=None)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 1, 0, 0, 1, 0],
    ...         "col2": [0, 1, 0, 1, 0, 1, 0],
    ...         "col3": [0, 0, 0, 0, 1, 1, 1],
    ...     }
    ... )
    >>> output = analyzer.analyze(frame)
    >>> output
    ColumnCooccurrenceOutput(shape=(7, 3), ignore_self=False)

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
    ) -> None:
        self._columns = tuple(columns) if columns is not None else None
        self._exclude_columns = exclude_columns

        check_column_missing_policy(missing_policy)
        self._missing_policy = missing_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(self.get_args())
        return f"{self.__class__.__qualname__}({args})"

    def analyze(self, frame: pl.DataFrame, lazy: bool = True) -> BaseOutput:
        self._check_input_columns(frame)
        output = self._analyze(frame)
        if not lazy:
            output = output.compute()
        return output

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self.get_args(), other.get_args(), equal_nan=equal_nan)

    def find_columns(self, frame: pl.DataFrame) -> tuple[str, ...]:
        r"""Find the columns to transform.

        Args:
            frame: The input DataFrame. Sometimes the columns to
                transform are found by analyzing the input
                DataFrame.

        Returns:
            The columns to transform.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from arkas.analyzer import ColumnCooccurrenceAnalyzer
        >>> frame = pl.DataFrame(
        ...     {
        ...         "col1": [0, 1, 1, 0, 0, 1, 0],
        ...         "col2": [0, 1, 0, 1, 0, 1, 0],
        ...         "col3": [0, 0, 0, 0, 1, 1, 1],
        ...     }
        ... )
        >>> analyzer = ColumnCooccurrenceAnalyzer(columns=["col2", "col3"])
        >>> analyzer.find_columns(frame)
        ('col2', 'col3')
        >>> analyzer = ColumnCooccurrenceAnalyzer()
        >>> analyzer.find_columns(frame)
        ('col1', 'col2', 'col3')

        ```
        """
        cols = list(frame.columns if self._columns is None else self._columns)
        [cols.remove(col) for col in self._exclude_columns if col in cols]
        return tuple(cols)

    def find_common_columns(self, frame: pl.DataFrame) -> tuple[str, ...]:
        r"""Find the common columns between the DataFrame columns and the
        input columns.

        Args:
            frame: The input DataFrame. Sometimes the columns to
                transform are found by analyzing the input
                DataFrame.

        Returns:
            The common columns.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from arkas.analyzer import ColumnCooccurrenceAnalyzer
        >>> frame = pl.DataFrame(
        ...     {
        ...         "col1": [0, 1, 1, 0, 0, 1, 0],
        ...         "col2": [0, 1, 0, 1, 0, 1, 0],
        ...         "col3": [0, 0, 0, 0, 1, 1, 1],
        ...     }
        ... )
        >>> analyzer = ColumnCooccurrenceAnalyzer(columns=["col2", "col3", "col5"])
        >>> analyzer.find_common_columns(frame)
        ('col2', 'col3')
        >>> analyzer = ColumnCooccurrenceAnalyzer()
        >>> analyzer.find_common_columns(frame)
        ('col1', 'col2', 'col3')

        ```
        """
        return find_common_columns(frame, self.find_columns(frame))

    def find_missing_columns(self, frame: pl.DataFrame) -> tuple[str, ...]:
        r"""Find the missing columns.

        Args:
            frame: The input DataFrame. Sometimes the columns to
                transform are found by analyzing the input
                DataFrame.

        Returns:
            The missing columns.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from arkas.analyzer import ColumnCooccurrenceAnalyzer
        >>> frame = pl.DataFrame(
        ...     {
        ...         "col1": [0, 1, 1, 0, 0, 1, 0],
        ...         "col2": [0, 1, 0, 1, 0, 1, 0],
        ...         "col3": [0, 0, 0, 0, 1, 1, 1],
        ...     }
        ... )
        >>> analyzer = ColumnCooccurrenceAnalyzer(columns=["col2", "col3", "col5"])
        >>> analyzer.find_missing_columns(frame)
        ('col5',)
        >>> analyzer = ColumnCooccurrenceAnalyzer()
        >>> analyzer.find_missing_columns(frame)
        ()

        ```
        """
        return find_missing_columns(frame, self.find_columns(frame))

    def get_args(self) -> dict:
        r"""Get the arguments of the analyzer.

        Returns:
            The arguments.
        """
        return {
            "columns": self._columns,
            "exclude_columns": self._exclude_columns,
            "missing_policy": self._missing_policy,
        }

    def _check_input_columns(self, frame: pl.DataFrame) -> None:
        r"""Check if some input columns are missing.

        Args:
            frame: The input DataFrame to check.
        """
        check_missing_columns(
            frame_or_cols=frame,
            columns=self.find_columns(frame),
            missing_policy=self._missing_policy,
        )

    @abstractmethod
    def _analyze(self, frame: pl.DataFrame) -> BaseOutput:
        r"""Analyze the DataFrame.

        Args:
            frame: The DataFrame to analyze.

        Returns:
            The generated output.
        """



r"""Implement the classification accuracy metrics."""

from __future__ import annotations

__all__ = ["balanced_accuracy"]


from typing import TYPE_CHECKING

from sklearn import metrics

from arkas.metric.utils import contains_nan, preprocess_pred

if TYPE_CHECKING:
    import numpy as np


def balanced_accuracy(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the accuracy metrics.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import balanced_accuracy
    >>> balanced_accuracy(y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1]))
    {'balanced_accuracy': 1.0, 'count': 5}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    count = y_true.size
    acc = float("nan")
    if count > 0 and not y_true_nan and not y_pred_nan:
        acc = float(metrics.balanced_accuracy_score(y_true=y_true, y_pred=y_pred))
    return {f"{prefix}balanced_accuracy{suffix}": acc, f"{prefix}count{suffix}": count}



r"""Contain output exporters."""

from __future__ import annotations

__all__ = [
    "BaseExporter",
    "FigureExporter",
    "MetricExporter",
    "ReportExporter",
    "SequentialExporter",
    "is_exporter_config",
    "setup_exporter",
]

from arkas.exporter.base import BaseExporter, is_exporter_config, setup_exporter
from arkas.exporter.figure import FigureExporter
from arkas.exporter.metric import MetricExporter
from arkas.exporter.report import ReportExporter
from arkas.exporter.sequential import SequentialExporter



r"""Contain the base class to implement a plotter."""

from __future__ import annotations

__all__ = ["BasePlotter"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualNanHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

if TYPE_CHECKING:
    from coola.equality import EqualityConfig


class BasePlotter(ABC):
    r"""Define the base class to implement a plotter.

    Example usage:

    ```pycon

    >>> from arkas.plotter import Plotter
    >>> plotter = Plotter()
    >>> plotter
    Plotter(count=0)

    ```
    """

    @abstractmethod
    def compute(self) -> BasePlotter:
        r"""Compute the figures and return a new plotter.

        Returns:
            A new plotter with the computed figures.

        Example usage:

        ```pycon

        >>> from arkas.plotter import Plotter
        >>> plotter = Plotter({"fig": None})
        >>> plotter
        Plotter(count=1)
        >>> plotter2 = plotter.compute()
        >>> plotter2
        Plotter(count=1)

        ```
        """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two plotters are equal or not.

        Args:
            other: The other plotter to compare.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two plotters are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> from arkas.plotter import Plotter
        >>> plotter1 = Plotter()
        >>> plotter2 = Plotter()
        >>> plotter3 = Plotter({"fig": None})
        >>> plotter1.equal(plotter2)
        True
        >>> plotter1.equal(plotter3)
        False

        ```
        """

    @abstractmethod
    def plot(self, prefix: str = "", suffix: str = "") -> dict:
        r"""Generate the figures.

        Args:
            prefix: The key prefix in the returned dictionary.
            suffix: The key suffix in the returned dictionary.

        Returns:
            A dictionary with the generated figures.

        Example usage:

        ```pycon

        >>> from arkas.plotter import Plotter
        >>> plotter = Plotter()
        >>> plotter.plot()
        {}

        ```
        """


class PlotterEqualityComparator(BaseEqualityComparator[BasePlotter]):
    r"""Implement an equality comparator for ``BasePlotter`` objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualNanHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> PlotterEqualityComparator:
        return self.__class__()

    def equal(self, actual: BasePlotter, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BasePlotter):  # pragma: no cover
    EqualityTester.add_comparator(BasePlotter, PlotterEqualityComparator())



r"""Implement the Area Under the Receiver Operating Characteristic Curve
(ROC AUC) result."""

from __future__ import annotations

__all__ = [
    "BaseRocAucResult",
    "BinaryRocAucResult",
    "MulticlassRocAucResult",
    "MultilabelRocAucResult",
]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric.classification.roc_auc import (
    binary_roc_auc,
    multiclass_roc_auc,
    multilabel_roc_auc,
)
from arkas.metric.utils import check_nan_policy, check_same_shape_score
from arkas.result.base import BaseResult

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    import numpy as np


class BaseRocAucResult(BaseResult):
    r"""Implement the base class to implement the Area Under the Receiver
    Operating Characteristic Curve (ROC AUC) results.

    Args:
        y_true: The ground truth target labels.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryRocAucResult
    >>> result = BinaryRocAucResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_score=np.array([1, 0, 0, 1, 1])
    ... )
    >>> result
    BinaryRocAucResult(y_true=(5,), y_score=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'roc_auc': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        self._y_true = y_true
        self._y_score = y_score

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_score": self._y_score.shape,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_score(self) -> np.ndarray:
        return self._y_score

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_score, other.y_score, equal_nan=equal_nan)
            and self.nan_policy == other.nan_policy
        )


class BinaryRocAucResult(BaseRocAucResult):
    r"""Implement the Area Under the Receiver Operating Characteristic
    Curve (ROC AUC) result for binary labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, *)`` with ``0`` and
            ``1`` values.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, *)``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryRocAucResult
    >>> result = BinaryRocAucResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_score=np.array([2, -1, 0, 3, 1])
    ... )
    >>> result
    BinaryRocAucResult(y_true=(5,), y_score=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'roc_auc': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_score(y_true, y_score)
        super().__init__(y_true=y_true.ravel(), y_score=y_score.ravel(), nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return binary_roc_auc(
            y_true=self._y_true,
            y_score=self._y_score,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, plt.Figure]:
        return {}


class MulticlassRocAucResult(BaseRocAucResult):
    r"""Implement the Area Under the Receiver Operating Characteristic
    Curve (ROC AUC) result for multiclass labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)`` with ``0`` and
            ``1`` values.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MulticlassRocAucResult
    >>> result = MulticlassRocAucResult(
    ...     y_true=np.array([0, 0, 1, 1, 2, 2]),
    ...     y_score=np.array(
    ...         [
    ...             [0.7, 0.2, 0.1],
    ...             [0.4, 0.3, 0.3],
    ...             [0.1, 0.8, 0.1],
    ...             [0.2, 0.5, 0.3],
    ...             [0.3, 0.3, 0.4],
    ...             [0.1, 0.2, 0.7],
    ...         ]
    ...     ),
    ... )
    >>> result
    MulticlassRocAucResult(y_true=(6,), y_score=(6, 3), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 6,
     'macro_roc_auc': 1.0,
     'micro_roc_auc': 1.0,
     'roc_auc': array([1., 1., 1.]),
     'weighted_roc_auc': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        y_true = y_true.ravel()
        if y_true.shape[0] != y_score.shape[0]:
            msg = (
                f"'y_true' and 'y_score' have different first dimension: {y_true.shape} vs "
                f"{y_score.shape}"
            )
            raise RuntimeError(msg)
        super().__init__(y_true=y_true, y_score=y_score, nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return multiclass_roc_auc(
            y_true=self._y_true,
            y_score=self._y_score,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, plt.Figure]:
        return {}


class MultilabelRocAucResult(BaseRocAucResult):
    r"""Implement the Area Under the Receiver Operating Characteristic
    Curve (ROC AUC) result for multilabel labels.

    Args:
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples, n_classes)`` with ``0``
            and ``1`` values.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples, n_classes)``.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import MultilabelRocAucResult
    >>> result = MultilabelRocAucResult(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_score=np.array([[2, -1, 1], [-1, 1, -2], [0, 2, -3], [3, -2, 4], [1, -3, 5]]),
    ... )
    >>> result
    MultilabelRocAucResult(y_true=(5, 3), y_score=(5, 3), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5,
     'macro_roc_auc': 1.0,
     'micro_roc_auc': 1.0,
     'roc_auc': array([1., 1., 1.]),
     'weighted_roc_auc': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_score: np.ndarray,
        nan_policy: str = "propagate",
    ) -> None:
        check_same_shape_score(y_true, y_score)
        super().__init__(y_true=y_true, y_score=y_score, nan_policy=nan_policy)

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return multilabel_roc_auc(
            y_true=self._y_true,
            y_score=self._y_score,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, plt.Figure]:
        return {}



r"""Contain an accuracy evaluator."""

from __future__ import annotations

__all__ = ["BalancedAccuracyEvaluator"]

import logging
from typing import TYPE_CHECKING

from coola.utils.format import repr_mapping_line

from arkas.evaluator.lazy import BaseLazyEvaluator
from arkas.metric.utils import check_nan_policy
from arkas.result import BalancedAccuracyResult, Result
from arkas.utils.array import to_array

if TYPE_CHECKING:
    import polars as pl


logger = logging.getLogger(__name__)


class BalancedAccuracyEvaluator(BaseLazyEvaluator[BalancedAccuracyResult]):
    r"""Implement the accuracy evaluator.

    Args:
        y_true: The key or column name of the ground truth target
            labels.
        y_pred: The key or column name of the predicted labels.
        drop_nulls: If ``True``, the rows with null values in
            ``y_true`` or ``y_pred`` columns are dropped.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.evaluator import BalancedAccuracyEvaluator
    >>> evaluator = BalancedAccuracyEvaluator(y_true="target", y_pred="pred")
    >>> evaluator
    BalancedAccuracyEvaluator(y_true='target', y_pred='pred', drop_nulls=True, nan_policy='propagate')
    >>> data = pl.DataFrame({"pred": [3, 2, 0, 1, 0, 1], "target": [3, 2, 0, 1, 0, 1]})
    >>> result = evaluator.evaluate(data)
    >>> result
    BalancedAccuracyResult(y_true=(6,), y_pred=(6,), nan_policy='propagate')

    ```
    """

    def __init__(
        self,
        y_true: str,
        y_pred: str,
        drop_nulls: bool = True,
        nan_policy: str = "propagate",
    ) -> None:
        super().__init__(drop_nulls=drop_nulls)
        self._y_true = y_true
        self._y_pred = y_pred

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true,
                "y_pred": self._y_pred,
                "drop_nulls": self._drop_nulls,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    def evaluate(self, data: pl.DataFrame, lazy: bool = True) -> BalancedAccuracyResult | Result:
        logger.info(
            f"Evaluating the balanced accuracy | y_true={self._y_true!r} | y_pred={self._y_pred!r}"
            f" | drop_nulls={self._drop_nulls} | nan_policy={self._nan_policy!r}"
        )
        return self._evaluate(data, lazy)

    def _compute_result(self, data: pl.DataFrame) -> BalancedAccuracyResult:
        return BalancedAccuracyResult(
            y_true=to_array(data[self._y_true]).ravel(),
            y_pred=to_array(data[self._y_pred]).ravel(),
            nan_policy=self._nan_policy,
        )

    def _get_columns(self) -> tuple[str, ...]:
        return (self._y_true, self._y_pred)



r"""Implement the default binary classification result."""

from __future__ import annotations

__all__ = ["BinaryClassificationResult"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric.utils import check_nan_policy
from arkas.result.accuracy import AccuracyResult
from arkas.result.ap import BinaryAveragePrecisionResult
from arkas.result.balanced_accuracy import BalancedAccuracyResult
from arkas.result.base import BaseResult
from arkas.result.confmat import BinaryConfusionMatrixResult
from arkas.result.fbeta import BinaryFbetaScoreResult
from arkas.result.jaccard import BinaryJaccardResult
from arkas.result.precision import BinaryPrecisionResult
from arkas.result.recall import BinaryRecallResult
from arkas.result.roc_auc import BinaryRocAucResult
from arkas.result.sequential import SequentialResult

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np


class BinaryClassificationResult(BaseResult):
    r"""Implement the default binary classification result.

    Args:
        y_true: The ground truth target binary labels. This input must
            be an array of shape ``(n_samples,)`` where the values
            are ``0`` or ``1``.
        y_pred: The predicted binary labels. This input must be an
            array of shape ``(n_samples,)`` where the values are ``0``
            or ``1``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions.
        betas: The betas used to compute the F-beta scores.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import BinaryClassificationResult
    >>> result = BinaryClassificationResult(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_pred=np.array([1, 0, 0, 1, 1]),
    ...     y_score=np.array([2, -1, 0, 3, 1]),
    ... )
    >>> result
    BinaryClassificationResult(y_true=(5,), y_pred=(5,), y_score=(5,), betas=(1,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'accuracy': 1.0,
     'count_correct': 5,
     'count_incorrect': 0,
     'count': 5,
     'error': 0.0,
     'balanced_accuracy': 1.0,
     'confusion_matrix': array([[2, 0], [0, 3]]),
     'false_negative_rate': 0.0,
     'false_negative': 0,
     'false_positive_rate': 0.0,
     'false_positive': 0,
     'true_negative_rate': 1.0,
     'true_negative': 2,
     'true_positive_rate': 1.0,
     'true_positive': 3,
     'f1': 1.0,
     'jaccard': 1.0,
     'precision': 1.0,
     'recall': 1.0,
     'average_precision': 1.0,
     'roc_auc': 1.0}

    ```
    """

    def __init__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
        y_score: np.ndarray | None = None,
        betas: Sequence[float] = (1,),
        nan_policy: str = "propagate",
    ) -> None:
        self._y_true = y_true.ravel()
        self._y_pred = y_pred.ravel()
        self._y_score = None if y_score is None else y_score.ravel()
        self._betas = tuple(betas)

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

        results = [
            AccuracyResult(y_true=self._y_true, y_pred=self._y_pred, nan_policy=self._nan_policy),
            BalancedAccuracyResult(
                y_true=self._y_true, y_pred=self._y_pred, nan_policy=self._nan_policy
            ),
            BinaryConfusionMatrixResult(
                y_true=self._y_true, y_pred=self._y_pred, nan_policy=self._nan_policy
            ),
            BinaryFbetaScoreResult(
                y_true=self._y_true,
                y_pred=self._y_pred,
                betas=self._betas,
                nan_policy=self._nan_policy,
            ),
            BinaryJaccardResult(
                y_true=self._y_true, y_pred=self._y_pred, nan_policy=self._nan_policy
            ),
            BinaryPrecisionResult(
                y_true=self._y_true, y_pred=self._y_pred, nan_policy=self._nan_policy
            ),
            BinaryRecallResult(
                y_true=self._y_true, y_pred=self._y_pred, nan_policy=self._nan_policy
            ),
        ]
        if self._y_score is not None:
            results.extend(
                [
                    BinaryAveragePrecisionResult(
                        y_true=self._y_true, y_score=self._y_score, nan_policy=self._nan_policy
                    ),
                    BinaryRocAucResult(
                        y_true=self._y_true, y_score=self._y_score, nan_policy=self._nan_policy
                    ),
                ]
            )
        self._results = SequentialResult(results)

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "y_true": self._y_true.shape,
                "y_pred": self._y_pred.shape,
                "y_score": self._y_score.shape if self._y_score is not None else None,
                "betas": self._betas,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def y_true(self) -> np.ndarray:
        return self._y_true

    @property
    def y_pred(self) -> np.ndarray:
        return self._y_pred

    @property
    def y_score(self) -> np.ndarray | None:
        return self._y_score

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return self._results.compute_metrics(prefix=prefix, suffix=suffix)

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.y_true, other.y_true, equal_nan=equal_nan)
            and objects_are_equal(self.y_pred, other.y_pred, equal_nan=equal_nan)
            and objects_are_equal(self.y_score, other.y_score, equal_nan=equal_nan)
            and objects_are_equal(self._betas, other._betas, equal_nan=equal_nan)
            and self.nan_policy == other.nan_policy
        )

    def generate_figures(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return self._results.generate_figures(prefix=prefix, suffix=suffix)



r"""Implement some utility functions for ``numpy.ndarray``s."""

from __future__ import annotations

__all__ = ["filter_range", "nonnan", "rand_replace", "to_array"]

from typing import Any

import numpy as np
import polars as pl
from coola.utils.array import to_array as coola_to_array


def filter_range(array: np.ndarray, xmin: float, xmax: float) -> np.ndarray:
    r"""Filter in the values in a given range.

    Args:
        array: The input array.
        xmin: The lower bound of the range.
        xmax: The upper bound of the range.

    Returns:
        A 1-d array with only the values in the given range.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.utils.array import filter_range
    >>> out = filter_range(np.arange(10), xmin=-1, xmax=5)
    >>> out
    array([0, 1, 2, 3, 4, 5])

    ```
    """
    return np.extract(np.logical_and(xmin <= array, array <= xmax), array)


def nonnan(array: np.ndarray) -> np.ndarray:
    r"""Return the non-NaN values of an array.

    Args:
        array: The input array.

    Returns:
        A 1d array with the non-NaN values of the input array.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.utils.array import nonnan
    >>> nonnan(np.asarray([1, 2, float("nan"), 5, 6]))
    array([1., 2., 5., 6.])
    >>> nonnan(np.asarray([[1, 2, float("nan")], [4, 5, 6]]))
    array([1., 2., 4., 5., 6.])

    ```
    """
    mask = np.isnan(array)
    return array[~mask]


def rand_replace(
    arr: np.ndarray, value: Any, prob: float = 0.5, rng: np.random.Generator | None = None
) -> np.ndarray:
    r"""Return an array that contains the same values as the input array,
    excepts some values are replaced by ``value``.

    Args:
        arr: The array with the original values.
        value: The value used to replace existing values.
        prob: The probability of value replacement.
            If the value is ``0.2``, it means each value as 20% chance
            to be replaced.
        rng: The random number generator used to decide which values
            are replaced or not. If ``None``, the default random
            number generator is used.

    Returns:
        The generated array.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.utils.array import rand_replace
    >>> rng = np.random.default_rng(42)
    >>> out = rand_replace(np.arange(10), value=-1, prob=0.4, rng=rng)

    ```
    """
    if rng is None:
        rng = np.random.default_rng()
    mask = rng.choice([True, False], size=arr.shape, p=[prob, 1.0 - prob])
    return np.where(mask, value, arr)


def to_array(data: Any) -> np.ndarray:
    r"""Convert the input to a ``numpy.ndarray``.

    Args:
        data: The data to convert to a NumPy array.

    Returns:
        A NumPy array.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.utils.array import to_array
    >>> to_array([1, 2, 3, 4, 5])
    array([1, 2, 3, 4, 5])
    >>> to_array(pl.Series([1, 2, 3, 4, 5]))
    array([1, 2, 3, 4, 5])
    >>> to_array(pl.DataFrame({"col1": [1, 2, 3, 4, 5], "col2": [0, 1, 0, 1, 0]}))
    array([[1, 0], [2, 1], [3, 0], [4, 1], [5, 0]])

    ```
    """
    if isinstance(data, pl.Series):
        return data.to_numpy()
    if isinstance(data, pl.DataFrame):
        return data.to_numpy()
    return coola_to_array(data)



r"""Contain confusion matrix metrics."""

from __future__ import annotations

__all__ = [
    "binary_confusion_matrix",
    "confusion_matrix",
    "multiclass_confusion_matrix",
    "multilabel_confusion_matrix",
]


import numpy as np
from sklearn import metrics

from arkas.metric.classification.precision import find_label_type
from arkas.metric.utils import (
    check_label_type,
    contains_nan,
    preprocess_pred,
    preprocess_pred_multilabel,
)


def confusion_matrix(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    label_type: str = "auto",
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the confusion matrix metrics.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        label_type: The type of labels used to evaluate the metrics.
            The valid values are: ``'binary'``, ``'multiclass'``,
            and ``'multilabel'``. If ``'binary'`` or ``'multilabel'``,
            ``y_true`` values  must be ``0`` and ``1``.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import confusion_matrix
    >>> # binary
    >>> confusion_matrix(
    ...     y_true=np.array([1, 0, 0, 1, 1]),
    ...     y_pred=np.array([1, 0, 0, 1, 1]),
    ...     label_type="binary",
    ... )
    {'confusion_matrix': array([[2, 0], [0, 3]]),
     'count': 5,
     'false_negative_rate': 0.0,
     'false_negative': 0,
     'false_positive_rate': 0.0,
     'false_positive': 0,
     'true_negative_rate': 1.0,
     'true_negative': 2,
     'true_positive_rate': 1.0,
     'true_positive': 3}
    >>> # multiclass
    >>> confusion_matrix(
    ...     y_true=np.array([0, 1, 1, 2, 2, 2]),
    ...     y_pred=np.array([0, 1, 1, 2, 2, 2]),
    ...     label_type="multiclass",
    ... )
    {'confusion_matrix': array([[1, 0, 0], [0, 2, 0], [0, 0, 3]]), 'count': 6}
    >>> # multilabel
    >>> confusion_matrix(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     label_type="multilabel",
    ... )
    {'confusion_matrix': array([[[2, 0], [0, 3]],
                                [[3, 0], [0, 2]],
                                [[2, 0], [0, 3]]]),
     'count': 5}

    ```
    """
    check_label_type(label_type)
    if label_type == "auto":
        label_type = find_label_type(y_true=y_true, y_pred=y_pred)
    if label_type == "binary":
        return binary_confusion_matrix(
            y_true=y_true, y_pred=y_pred, prefix=prefix, suffix=suffix, nan_policy=nan_policy
        )
    if label_type == "multilabel":
        return multilabel_confusion_matrix(
            y_true=y_true, y_pred=y_pred, prefix=prefix, suffix=suffix, nan_policy=nan_policy
        )
    return multiclass_confusion_matrix(
        y_true=y_true, y_pred=y_pred, prefix=prefix, suffix=suffix, nan_policy=nan_policy
    )


def binary_confusion_matrix(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the confusion matrix metrics for binary labels.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import binary_confusion_matrix
    >>> binary_confusion_matrix(
    ...     y_true=np.array([1, 0, 0, 1, 1]), y_pred=np.array([1, 0, 0, 1, 1])
    ... )
    {'confusion_matrix': array([[2, 0], [0, 3]]),
     'count': 5,
     'false_negative_rate': 0.0,
     'false_negative': 0,
     'false_positive_rate': 0.0,
     'false_positive': 0,
     'true_negative_rate': 1.0,
     'true_negative': 2,
     'true_positive_rate': 1.0,
     'true_positive': 3}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    count = y_true.size
    if y_true_nan or y_pred_nan:
        confmat = np.array([[np.nan, np.nan], [np.nan, np.nan]])
    elif count > 0:
        confmat = metrics.confusion_matrix(y_true=y_true, y_pred=y_pred)
    else:
        confmat = np.zeros((2, 2), dtype=np.int64)
    tn, fp, fn, tp = confmat.ravel().tolist()
    neg = tn + fp
    pos = tp + fn
    return {
        f"{prefix}confusion_matrix{suffix}": confmat,
        f"{prefix}count{suffix}": count,
        f"{prefix}false_negative_rate{suffix}": fn / pos if pos > 0 else float("nan"),
        f"{prefix}false_negative{suffix}": fn,
        f"{prefix}false_positive_rate{suffix}": fp / neg if neg > 0 else float("nan"),
        f"{prefix}false_positive{suffix}": fp,
        f"{prefix}true_negative_rate{suffix}": tn / neg if neg > 0 else float("nan"),
        f"{prefix}true_negative{suffix}": tn,
        f"{prefix}true_positive_rate{suffix}": tp / pos if pos > 0 else float("nan"),
        f"{prefix}true_positive{suffix}": tp,
    }


def multiclass_confusion_matrix(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the confusion matrix metrics for multiclass labels.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import multiclass_confusion_matrix
    >>> multiclass_confusion_matrix(
    ...     y_true=np.array([0, 1, 1, 2, 2, 2]), y_pred=np.array([0, 1, 1, 2, 2, 2])
    ... )
    {'confusion_matrix': array([[1, 0, 0], [0, 2, 0], [0, 0, 3]]), 'count': 6}

    ```
    """
    y_true, y_pred = preprocess_pred(
        y_true=y_true.ravel(), y_pred=y_pred.ravel(), drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")
    if y_true_nan or y_pred_nan:
        confmat = np.zeros((0, 0), dtype=np.int64)
    else:
        confmat = metrics.confusion_matrix(y_true=y_true, y_pred=y_pred)
    return {
        f"{prefix}confusion_matrix{suffix}": confmat,
        f"{prefix}count{suffix}": y_true.size,
    }


def multilabel_confusion_matrix(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    *,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float | np.ndarray]:
    r"""Return the confusion matrix metrics for multilabel labels.

    Args:
        y_true: The ground truth target labels.
        y_pred: The predicted labels.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import multilabel_confusion_matrix
    >>> multilabel_confusion_matrix(
    ...     y_true=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ...     y_pred=np.array([[1, 0, 1], [0, 1, 0], [0, 1, 0], [1, 0, 1], [1, 0, 1]]),
    ... )
    {'confusion_matrix': array([[[2, 0], [0, 3]],
                                [[3, 0], [0, 2]],
                                [[2, 0], [0, 3]]]),
     'count': 5}

    ```
    """
    y_true, y_pred = preprocess_pred_multilabel(y_true, y_pred, drop_nan=nan_policy == "omit")
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_pred_nan = contains_nan(arr=y_pred, nan_policy=nan_policy, name="'y_pred'")

    count = y_true.shape[0]
    if count > 0 and not y_true_nan and not y_pred_nan:
        if y_true.shape[1] > 1:
            confmat = metrics.multilabel_confusion_matrix(y_true=y_true, y_pred=y_pred)
        else:
            confmat = metrics.confusion_matrix(y_true=y_true, y_pred=y_pred).reshape(1, 2, 2)
    else:
        confmat = np.zeros((0, 0, 0), dtype=np.int64)
    return {f"{prefix}confusion_matrix{suffix}": confmat, f"{prefix}count{suffix}": count}



r"""Contain Receiver Operating Characteristic Curve (ROC) plotting
functionalities."""

from __future__ import annotations

__all__ = ["binary_roc_curve"]

from typing import TYPE_CHECKING, Any

from sklearn.metrics import RocCurveDisplay

from arkas.metric.utils import preprocess_score_binary

if TYPE_CHECKING:
    import numpy as np
    from matplotlib.axes import Axes


def binary_roc_curve(ax: Axes, y_true: np.ndarray, y_score: np.ndarray, **kwargs: Any) -> None:
    r"""Plot the Receiver Operating Characteristic Curve (ROC) for binary
    labels.

    Args:
        ax: The axes of the matplotlib figure to update.
        y_true: The ground truth target labels. This input must
            be an array of shape ``(n_samples,)``.
        y_score: The target scores, can either be probability
            estimates of the positive class, confidence values,
            or non-thresholded measure of decisions. This input must
            be an array of shape ``(n_samples,)``.
        **kwargs: Arbitrary keyword arguments that are passed to
            ``RocCurveDisplay.from_predictions``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from arkas.plot import binary_roc_curve
    >>> fig, ax = plt.subplots()
    >>> binary_roc_curve(
    ...     ax=ax, y_true=np.array([1, 0, 0, 1, 1]), y_score=np.array([2, -1, 0, 3, 1])
    ... )

    ```
    """
    y_true, y_score = preprocess_score_binary(
        y_true=y_true.ravel(), y_score=y_score.ravel(), drop_nan=True
    )
    RocCurveDisplay.from_predictions(y_true=y_true, y_pred=y_score, ax=ax, **kwargs)



r"""Contain plotting functionalities."""

from __future__ import annotations

__all__ = ["binary_precision_recall_curve", "binary_roc_curve"]

from arkas.plot.pr import binary_precision_recall_curve
from arkas.plot.roc import binary_roc_curve



r"""Contain the base class to implement a runner."""

from __future__ import annotations

__all__ = ["BaseRunner", "is_runner_config", "setup_runner"]

import logging
from abc import ABC, abstractmethod
from typing import Any

from objectory import AbstractFactory
from objectory.utils import is_object_config

logger = logging.getLogger(__name__)


class BaseRunner(ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a runner.

    Example usage:

    ```pycon

    >>> import tempfile
    >>> import polars as pl
    >>> from pathlib import Path
    >>> from grizz.ingestor import Ingestor
    >>> from grizz.transformer import SequentialTransformer
    >>> from arkas.analyzer import AccuracyAnalyzer
    >>> from arkas.exporter import MetricExporter
    >>> from arkas.runner import AnalysisRunner
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     runner = AnalysisRunner(
    ...         ingestor=Ingestor(
    ...             pl.DataFrame(
    ...                 {
    ...                     "pred": [3, 2, 0, 1, 0],
    ...                     "target": [3, 2, 0, 1, 0],
    ...                 }
    ...             )
    ...         ),
    ...         transformer=SequentialTransformer(transformers=[]),
    ...         analyzer=AccuracyAnalyzer(y_true="target", y_pred="pred"),
    ...         exporter=MetricExporter(Path(tmpdir).joinpath("metrics.pkl")),
    ...     )
    ...     print(runner)
    ...     runner.run()
    ...
    AnalysisRunner(
      (ingestor): Ingestor(shape=(5, 2))
      (transformer): SequentialTransformer()
      (analyzer): AccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')
      (exporter): MetricExporter(
          (path): .../metrics.pkl
          (saver): PickleSaver(protocol=5)
          (exist_ok): False
          (show_metrics): True
        )
      (lazy): True
    )

    ```
    """

    @abstractmethod
    def run(self) -> Any:
        r"""Execute the logic of the runner.

        Returns:
            Any artifact of the runner

        Example usage:

        ```pycon

        >>> import tempfile
        >>> import polars as pl
        >>> from pathlib import Path
        >>> from grizz.ingestor import Ingestor
        >>> from grizz.transformer import SequentialTransformer
        >>> from arkas.analyzer import AccuracyAnalyzer
        >>> from arkas.exporter import MetricExporter
        >>> from arkas.runner import AnalysisRunner
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     runner = AnalysisRunner(
        ...         ingestor=Ingestor(
        ...             pl.DataFrame(
        ...                 {
        ...                     "pred": [3, 2, 0, 1, 0],
        ...                     "target": [3, 2, 0, 1, 0],
        ...                 }
        ...             )
        ...         ),
        ...         transformer=SequentialTransformer(transformers=[]),
        ...         analyzer=AccuracyAnalyzer(y_true="target", y_pred="pred"),
        ...         exporter=MetricExporter(Path(tmpdir).joinpath("metrics.pkl")),
        ...     )
        ...     runner.run()
        ...

        ```
        """


def is_runner_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseRunner``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseRunner`` object.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.runner import is_runner_config
    >>> is_runner_config(
    ...     {
    ...         "_target_": "arkas.runner.AnalysisRunner",
    ...         "ingestor": {
    ...             "_target_": "grizz.ingestor.Ingestor",
    ...             "frame": pl.DataFrame(
    ...                 {
    ...                     "pred": [3, 2, 0, 1, 0],
    ...                     "target": [3, 2, 0, 1, 0],
    ...                 }
    ...             ),
    ...         },
    ...         "transformer": {"_target_": "grizz.transformer.DropDuplicate"},
    ...         "analyzer": {
    ...             "_target_": "arkas.analyzer.AccuracyAnalyzer",
    ...             "y_true": "target",
    ...             "y_pred": "pred",
    ...         },
    ...         "exporter": {
    ...             "_target_": "arkas.exporter.MetricExporter",
    ...             "path": "/path/to/data.csv",
    ...         },
    ...     }
    ... )
    True

    ```
    """
    return is_object_config(config, BaseRunner)


def setup_runner(
    runner: BaseRunner | dict,
) -> BaseRunner:
    r"""Set up a runner.

    The runner is instantiated from its configuration
    by using the ``BaseRunner`` factory function.

    Args:
        runner: Specifies a runner or its configuration.

    Returns:
        An instantiated runner.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from arkas.runner import setup_runner
    >>> runner = setup_runner(
    ...     {
    ...         "_target_": "arkas.runner.AnalysisRunner",
    ...         "ingestor": {
    ...             "_target_": "grizz.ingestor.Ingestor",
    ...             "frame": pl.DataFrame(
    ...                 {
    ...                     "pred": [3, 2, 0, 1, 0],
    ...                     "target": [3, 2, 0, 1, 0],
    ...                 }
    ...             ),
    ...         },
    ...         "transformer": {"_target_": "grizz.transformer.DropDuplicate"},
    ...         "analyzer": {
    ...             "_target_": "arkas.analyzer.AccuracyAnalyzer",
    ...             "y_true": "target",
    ...             "y_pred": "pred",
    ...         },
    ...         "exporter": {
    ...             "_target_": "arkas.exporter.MetricExporter",
    ...             "path": "/path/to/data.csv",
    ...         },
    ...     }
    ... )
    >>> runner
    AnalysisRunner(
      (ingestor): Ingestor(shape=(5, 2))
      (transformer): DropDuplicateTransformer(columns=None, exclude_columns=(), missing_policy='raise')
      (analyzer): AccuracyAnalyzer(y_true='target', y_pred='pred', drop_nulls=True, missing_policy='raise', nan_policy='propagate')
      (exporter): MetricExporter(
          (path): /path/to/data.csv
          (saver): PickleSaver(protocol=5)
          (exist_ok): False
          (show_metrics): True
        )
      (lazy): True
    )

    ```
    """
    if isinstance(runner, dict):
        logger.info("Initializing a runner from its configuration... ")
        runner = BaseRunner.factory(**runner)
    if not isinstance(runner, BaseRunner):
        logger.warning(f"runner is not a `BaseRunner` (received: {type(runner)})")
    return runner



r"""Implement the Normalized Discounted Cumulative Gain (NDCG)
metrics."""

from __future__ import annotations

__all__ = ["ndcg"]


from typing import TYPE_CHECKING

from sklearn.metrics import ndcg_score

from arkas.metric.utils import (
    check_array_ndim,
    contains_nan,
    preprocess_score_multilabel,
)

if TYPE_CHECKING:
    import numpy as np


def ndcg(
    y_true: np.ndarray,
    y_score: np.ndarray,
    *,
    k: int | None = None,
    prefix: str = "",
    suffix: str = "",
    nan_policy: str = "propagate",
) -> dict[str, float]:
    r"""Return the Normalized Discounted Cumulative Gain (NDCG) metrics.

    Args:
        y_true: The ground truth target targets of multilabel
            classification, or true scores of entities to be ranked.
            Negative values in y_true may result in an output that is
            not between 0 and 1. This input must be an array of shape
            ``(n_samples, n_labels)``.
        y_score: The predicted scores, can either be probability
            estimates, confidence values, or non-thresholded measure
            of decisions. This input must be an array of shape
            ``(n_samples, n_labels)``.
        k: Only consider the highest ``k`` scores in the ranking.
            If ``None``, use all outputs.
        prefix: The key prefix in the returned dictionary.
        suffix: The key suffix in the returned dictionary.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Returns:
        The computed metrics.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.metric import ndcg
    >>> ndcg(
    ...     y_true=np.array([[1, 0, 0], [1, 2, 0], [1, 1, 2], [0, 0, 1]]),
    ...     y_score=np.array(
    ...         [[2.0, 1.0, 0.0], [0.0, 1.0, -1.0], [0.0, 0.0, 1.0], [1.0, 2.0, 3.0]]
    ...     ),
    ... )
    {'count': 4, 'ndcg': 1.0}

    ```
    """
    check_array_ndim(y_true, ndim=2)
    y_true, y_score = preprocess_score_multilabel(
        y_true=y_true, y_score=y_score, drop_nan=nan_policy == "omit"
    )
    y_true_nan = contains_nan(arr=y_true, nan_policy=nan_policy, name="'y_true'")
    y_score_nan = contains_nan(arr=y_score, nan_policy=nan_policy, name="'y_score'")

    n_samples = y_true.shape[0]
    score = float("nan")
    if n_samples > 0 and not y_true_nan and not y_score_nan:
        score = float(ndcg_score(y_true=y_true, y_score=y_score, k=k))
    return {f"{prefix}count{suffix}": n_samples, f"{prefix}ndcg{suffix}": score}



"""Contain the base class to implement a figure."""

from __future__ import annotations

__all__ = ["BaseFigure", "BaseFigureConfig"]

import sys
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualNanHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester

if TYPE_CHECKING:
    from coola.equality import EqualityConfig

if sys.version_info >= (3, 11):
    from typing import Self
else:  # pragma: no cover
    from typing_extensions import (
        Self,  # use backport because it was added in python 3.11
    )


class BaseFigure(ABC):
    r"""Define the base class to implement a figure.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from arkas.figure import MatplotlibFigure
    >>> fig = MatplotlibFigure(plt.subplots()[0])
    >>> fig
    MatplotlibFigure(reactive=True)

    ```
    """

    @abstractmethod
    def close(self) -> None:
        r"""Close the figure.

        Example usage:

        ```pycon

        >>> from matplotlib import pyplot as plt
        >>> from arkas.figure import MatplotlibFigure
        >>> fig = MatplotlibFigure(plt.subplots()[0])
        >>> fig.close()

        ```
        """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two figures are equal or not.

        Args:
            other: The other object to compare with.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two objects are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> from matplotlib import pyplot as plt
        >>> from arkas.figure import MatplotlibFigure
        >>> fig = plt.subplots()[0]
        >>> fig1 = MatplotlibFigure(fig)
        >>> fig2 = MatplotlibFigure(fig)
        >>> fig3 = MatplotlibFigure(fig, reactive=False)
        >>> fig1.equal(fig2)
        True
        >>> fig1.equal(fig3)
        False

        ```
        """

    @abstractmethod
    def set_reactive(self, reactive: bool) -> Self:
        r"""Set the figure reactive mode.

        Args:
            reactive: The reactive mode.

        Returns:
            A new figure object with the right reactive mode.

        Example usage:

        ```pycon

        >>> from matplotlib import pyplot as plt
        >>> from arkas.figure import MatplotlibFigure
        >>> fig = MatplotlibFigure(plt.subplots()[0])
        >>> fig
        MatplotlibFigure(reactive=True)
        >>> fig2 = fig.set_reactive(False)
        >>> fig2
        MatplotlibFigure(reactive=False)

        ```
        """

    @abstractmethod
    def to_html(self) -> str:
        r"""Export the figure to a HTML code.

        Returns:
            The HTML code of the figure.

        Example usage:

        ```pycon

        >>> from matplotlib import pyplot as plt
        >>> from arkas.figure import MatplotlibFigure
        >>> fig = MatplotlibFigure(plt.subplots()[0])
        >>> html = fig.to_html()

        ```
        """


class BaseFigureConfig(ABC):
    r"""Define the base class to implement a figure config.

    Example usage:

    ```pycon

    >>> from arkas.figure import MatplotlibFigureConfig
    >>> config = MatplotlibFigureConfig()
    >>> config
    MatplotlibFigureConfig(color_norm=None)

    ```
    """

    @classmethod
    @abstractmethod
    def backend(cls) -> str:
        r"""Return the backend to generate the figure.

        Example usage:

        ```pycon

        >>> from arkas.figure import MatplotlibFigureConfig
        >>> backend = MatplotlibFigureConfig.backend()
        >>> backend
        matplotlib

        ```
        """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two configs are equal or not.

        Args:
            other: The other object to compare with.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two objects are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> from arkas.figure import MatplotlibFigureConfig
        >>> config1 = MatplotlibFigureConfig(dpi=300)
        >>> config2 = MatplotlibFigureConfig(dpi=300)
        >>> config3 = MatplotlibFigureConfig()
        >>> config1.equal(config2)
        True
        >>> config1.equal(config3)
        False

        ```
        """

    @abstractmethod
    def get_args(self) -> dict:
        r"""Get the config arguments.

        Returns:
            The config arguments.

        Example usage:

        ```pycon

        >>> from arkas.figure import MatplotlibFigureConfig
        >>> config = MatplotlibFigureConfig(dpi=300)
        >>> config.get_args()
        {'dpi': 300}

        ```
        """


class FigureEqualityComparator(BaseEqualityComparator[BaseFigure]):
    r"""Implement an equality comparator for ``BaseFigure`` and
    ``BaseFigureConfig`` objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualNanHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> FigureEqualityComparator:
        return self.__class__()

    def equal(
        self, actual: BaseFigure | BaseFigureConfig, expected: Any, config: EqualityConfig
    ) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BaseFigure):  # pragma: no cover
    EqualityTester.add_comparator(BaseFigure, FigureEqualityComparator())
    EqualityTester.add_comparator(BaseFigureConfig, FigureEqualityComparator())



r"""Contain utility functions for figure's ticks."""

from __future__ import annotations

__all__ = ["readable_xticklabels"]

import math
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from matplotlib.axes import Axes


def readable_xticklabels(
    ax: Axes,
    max_num_xticks: int = 100,
    xticklabel_max_len: int = 20,
    xticklabel_min: int = 10,
) -> None:
    r"""Update the tick labels to make them easier to read, in particular
    if the tick labels are dense.

    Args:
        ax: The figure axes to update.
        max_num_xticks: The maximum number of ticks to show in the
            figure.
        xticklabel_max_len: If a tick label has a length greater than
            this value, the tick labels are rotated vertically.
        xticklabel_min: If the number of ticks is lower than this
            number the tick labels are rotated vertically.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from arkas.plot.utils import readable_xticklabels
    >>> fig, ax = plt.subplots()
    >>> ax.hist(np.arange(10), bins=10)
    >>> readable_xticklabels(ax)

    ```
    """
    xticks = ax.get_xticks()
    if len(xticks) > max_num_xticks:
        n = math.ceil(len(xticks) / max_num_xticks)
        xticks = xticks[::n]
        ax.set_xticks(xticks)
    if len(xticks) > xticklabel_min or any(
        len(str(label)) > xticklabel_max_len for label in ax.get_xticklabels()
    ):
        ax.tick_params(axis="x", labelrotation=90)


def readable_yticklabels(ax: Axes, max_num_yticks: int = 100) -> None:
    r"""Update the tick labels to make them easier to read, in particular
    if the tick labels are dense.

    Args:
        ax: The figure axes to update.
        max_num_yticks: The maximum number of ticks to show in the
            figure.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from arkas.plot.utils import readable_yticklabels
    >>> fig, ax = plt.subplots()
    >>> ax.hist(np.arange(10), bins=10)
    >>> readable_yticklabels(ax)

    ```
    """
    yticks = ax.get_yticks()
    if len(yticks) > max_num_yticks:
        n = math.ceil(len(yticks) / max_num_yticks)
        yticks = yticks[::n]
        ax.set_yticks(yticks)



r"""Implement a simple content output."""

from __future__ import annotations

__all__ = ["ContentOutput"]


from arkas.content.vanilla import ContentGenerator
from arkas.evaluator2.vanilla import Evaluator
from arkas.output.vanilla import Output
from arkas.plotter.vanilla import Plotter


class ContentOutput(Output):
    r"""Implement a simple content output.

    Args:
        content: The HTML content.

    Example usage:

    ```pycon

    >>> from arkas.output import ContentOutput
    >>> output = ContentOutput("meow")
    >>> output
    ContentOutput(
      (content): ContentGenerator()
      (evaluator): Evaluator(count=0)
      (plotter): Plotter(count=0)
    )
    >>> output.get_content_generator()
    ContentGenerator()
    >>> output.get_evaluator()
    Evaluator(count=0)
    >>> output.get_plotter()
    Plotter(count=0)

    ```
    """

    def __init__(self, content: str) -> None:
        super().__init__(
            content=ContentGenerator(content), evaluator=Evaluator(), plotter=Plotter()
        )



r"""Contain utility functions to generate sections."""

from __future__ import annotations

__all__ = [
    "GO_TO_TOP",
    "render_toc",
    "tags2id",
    "tags2title",
    "valid_h_tag",
]

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

GO_TO_TOP = '<a href="#">Go to top</a>'


def tags2id(tags: Sequence[str]) -> str:
    r"""Convert a sequence of tags to a string that can be used as ID in
    a HTML file.

    Args:
        tags: The sequence of tags.

    Returns:
        The generated ID from the tags.

    Example usage:

    ```pycon

    >>> from arkas.utils.html import tags2id
    >>> out = tags2id(["super", "meow"])
    >>> out
    super-meow

    ```
    """
    return "-".join(tags).replace(" ", "-").lower()


def tags2title(tags: Sequence[str]) -> str:
    r"""Convert a sequence of tags to a string that can be used as title.

    Args:
        tags: The sequence of tags.

    Returns:
        The generated title from the tags.

    Example usage:

    ```pycon

    >>> from arkas.utils.html import tags2title
    >>> out = tags2title(["super", "meow"])
    >>> out
    meow | super

    ```
    """
    return " | ".join(tags[::-1])


def valid_h_tag(index: int) -> int:
    r"""Return a valid number of a h HTML tag.

    Args:
        index: The original value.

    Returns:
        A valid value.

    Example usage:

    ```pycon

    >>> from arkas.utils.html import valid_h_tag
    >>> out = valid_h_tag(4)
    >>> out
    4

    ```
    """
    return max(1, min(6, index))


def render_toc(
    number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
) -> str:
    r"""Return the HTML table of content (TOC) associated to the section.

    Args:
        number: The section number associated to the section.
        tags: The tags associated to the section.
        depth: The depth in the report.
        max_depth: The maximum depth to generate in the TOC.

    Returns:
        The HTML table of content associated to the section.

    Example usage:

    ```pycon

    >>> from arkas.utils.html import render_toc
    >>> out = render_toc(number="2.0", tags=["super", "meow"])
    >>> out
    <li><a href="#super-meow">2.0 meow</a></li>

    ```
    """
    if depth >= max_depth:
        return ""
    tag = tags[-1] if tags else ""
    return f'<li><a href="#{tags2id(tags)}">{number} {tag}</a></li>'



r"""Implement the energy distance between two 1D distributions
result."""

from __future__ import annotations

__all__ = ["EnergyDistanceResult"]

from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils.format import repr_mapping_line

from arkas.metric import energy_distance
from arkas.metric.utils import check_nan_policy, check_same_shape
from arkas.result.base import BaseResult

if TYPE_CHECKING:
    import numpy as np


class EnergyDistanceResult(BaseResult):
    r"""Implement the energy distance between two 1D distributions
    result.

    Args:
        u_values: The values observed in the (empirical) distribution.
        v_values: The values observed in the (empirical) distribution.
        nan_policy: The policy on how to handle NaN values in the input
            arrays. The following options are available: ``'omit'``,
            ``'propagate'``, and ``'raise'``.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.result import EnergyDistanceResult
    >>> result = EnergyDistanceResult(
    ...     u_values=np.array([1, 2, 3, 4, 5]), v_values=np.array([1, 2, 3, 4, 5])
    ... )
    >>> result
    EnergyDistanceResult(u_values=(5,), v_values=(5,), nan_policy='propagate')
    >>> result.compute_metrics()
    {'count': 5, 'energy_distance': 0.0}

    ```
    """

    def __init__(
        self, u_values: np.ndarray, v_values: np.ndarray, nan_policy: str = "propagate"
    ) -> None:
        self._u_values = u_values.ravel()
        self._v_values = v_values.ravel()
        check_same_shape([self._u_values, self._v_values])

        check_nan_policy(nan_policy)
        self._nan_policy = nan_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "u_values": self._u_values.shape,
                "v_values": self._v_values.shape,
                "nan_policy": self._nan_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    @property
    def nan_policy(self) -> str:
        return self._nan_policy

    @property
    def u_values(self) -> np.ndarray:
        return self._u_values

    @property
    def v_values(self) -> np.ndarray:
        return self._v_values

    def compute_metrics(self, prefix: str = "", suffix: str = "") -> dict[str, float]:
        return energy_distance(
            u_values=self._u_values,
            v_values=self._v_values,
            prefix=prefix,
            suffix=suffix,
            nan_policy=self._nan_policy,
        )

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return (
            objects_are_equal(self.u_values, other.u_values, equal_nan=equal_nan)
            and objects_are_equal(self.v_values, other.v_values, equal_nan=equal_nan)
            and self.nan_policy == other.nan_policy
        )

    def generate_figures(
        self, prefix: str = "", suffix: str = ""  # noqa: ARG002
    ) -> dict[str, float]:
        return {}



r"""Contain the definition of a figure creator and a registry."""

from __future__ import annotations

__all__ = ["FigureCreatorRegistry"]

from typing import Any, Generic, TypeVar

from coola import objects_are_equal
from coola.utils import str_indent, str_mapping

T = TypeVar("T")


class FigureCreatorRegistry(Generic[T]):
    """Implement figure creator registry.

    Args:
        registry: The initial registry with the figure creators.
    """

    def __init__(self, registry: dict[str, T] | None = None) -> None:
        self._registry = registry or {}

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(self._registry))}\n)"

    def add_creator(self, backend: str, creator: T, exist_ok: bool = False) -> None:
        r"""Add a figure creator for a given backend.

        Args:
            backend: The backend for this test.
            creator: The creator used to test the figure
                of the specified type.
            exist_ok: If ``False``, ``RuntimeError`` is raised if the
                backend already exists. This parameter should be
                set to ``True`` to overwrite the creator for a type.

        Raises:
            RuntimeError: if an creator is already registered for the
                backend and ``exist_ok=False``.

        Example usage:

        ```pycon

        >>> from arkas.figure.creator import FigureCreatorRegistry
        >>> registry = FigureCreatorRegistry()

        ```
        """
        if backend in self._registry and not exist_ok:
            msg = (
                f"A figure creator ({self._registry[backend]}) is already registered for the "
                f"backend {backend!r}. Please use `exist_ok=True` if you want to overwrite the "
                "creator for this backend"
            )
            raise RuntimeError(msg)
        self._registry[backend] = creator

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two registries are equal or not.

        Args:
            other: The other object to compare with.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two objects are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> from arkas.figure.creator import FigureCreatorRegistry
        >>> registry = FigureCreatorRegistry()

        ```
        """
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._registry, other._registry, equal_nan=equal_nan)

    def has_creator(self, backend: str) -> bool:
        r"""Indicate if a figure creator is registered for the given
        backend.

        Args:
            backend: The backend to check.

        Returns:
            ``True`` if a figure creator is registered,
                otherwise ``False``.

        Example usage:

        ```pycon

        >>> from arkas.figure.creator import FigureCreatorRegistry
        >>> registry = FigureCreatorRegistry()
        >>> registry.has_creator("missing")
        False

        ```
        """
        return backend in self._registry

    def find_creator(self, backend: str) -> T:
        r"""Find the figure creator associated to a backend.

        Args:
            backend: The backend.

        Returns:
            The figure creator associated to the backend.

        Raises:
            ValueError: if the backend is missing.

        Example usage:

        ```pycon

        >>> from arkas.figure.creator import FigureCreatorRegistry
        >>> registry = FigureCreatorRegistry()

        ```
        """
        creator = self._registry.get(backend, None)
        if creator is not None:
            return creator
        msg = f"Incorrect backend: {backend!r}"
        raise ValueError(msg)



r"""Contain an evaluator that evaluates a mapping of evaluators."""

from __future__ import annotations

__all__ = ["EvaluatorDict"]

import logging
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal
from coola.utils import repr_indent, repr_mapping

from arkas.evaluator2.base import BaseEvaluator
from arkas.evaluator2.vanilla import Evaluator

if TYPE_CHECKING:
    from collections.abc import Hashable, Mapping

logger = logging.getLogger(__name__)


class EvaluatorDict(BaseEvaluator):
    r"""Implement an evaluator that sequentially evaluates a mapping of
    evaluators.

    Args:
        evaluators: The mapping of evaluators to evaluate.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from arkas.state import AccuracyState
    >>> from arkas.evaluator2 import EvaluatorDict, AccuracyEvaluator, Evaluator
    >>> evaluator = EvaluatorDict(
    ...     {
    ...         "one": AccuracyEvaluator(
    ...             AccuracyState(
    ...                 y_true=np.array([1, 0, 0, 1, 1]),
    ...                 y_pred=np.array([1, 0, 0, 1, 1]),
    ...                 y_true_name="target",
    ...                 y_pred_name="pred",
    ...             )
    ...         ),
    ...         "two": Evaluator({"accuracy": 1.0, "total": 42}),
    ...     }
    ... )
    >>> evaluator
    EvaluatorDict(
      (one): AccuracyEvaluator(
          (state): AccuracyState(y_true=(5,), y_pred=(5,), y_true_name='target', y_pred_name='pred')
          (nan_policy): propagate
        )
      (two): Evaluator(count=2)
    )
    >>> metrics = evaluator.evaluate()
    >>> metrics
    {'one': {'accuracy': 1.0, 'count_correct': 5, 'count_incorrect': 0, 'count': 5, 'error': 0.0},
     'two': {'accuracy': 1.0, 'total': 42}}

    ```
    """

    def __init__(self, evaluators: Mapping[Hashable, BaseEvaluator]) -> None:
        self._evaluators = evaluators

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping(self._evaluators))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute(self) -> Evaluator:
        return Evaluator(metrics=self.evaluate())

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._evaluators, other._evaluators, equal_nan=equal_nan)

    def evaluate(self, prefix: str = "", suffix: str = "") -> dict:
        return {
            key: evaluator.evaluate(prefix=prefix, suffix=suffix)
            for key, evaluator in self._evaluators.items()
        }



r"""Contain functions to compute metrics."""

from __future__ import annotations

__all__ = [
    "accuracy",
    "average_precision",
    "balanced_accuracy",
    "binary_average_precision",
    "binary_confusion_matrix",
    "binary_fbeta_score",
    "binary_jaccard",
    "binary_precision",
    "binary_recall",
    "binary_roc_auc",
    "binary_top_k_accuracy",
    "confusion_matrix",
    "energy_distance",
    "fbeta_score",
    "jaccard",
    "jensen_shannon_divergence",
    "kl_div",
    "mean_absolute_error",
    "mean_absolute_percentage_error",
    "mean_squared_error",
    "mean_squared_log_error",
    "mean_tweedie_deviance",
    "median_absolute_error",
    "multiclass_average_precision",
    "multiclass_confusion_matrix",
    "multiclass_fbeta_score",
    "multiclass_jaccard",
    "multiclass_precision",
    "multiclass_recall",
    "multiclass_roc_auc",
    "multiclass_top_k_accuracy",
    "multilabel_average_precision",
    "multilabel_confusion_matrix",
    "multilabel_fbeta_score",
    "multilabel_jaccard",
    "multilabel_precision",
    "multilabel_recall",
    "multilabel_roc_auc",
    "ndcg",
    "pearsonr",
    "precision",
    "r2_score",
    "recall",
    "regression_errors",
    "roc_auc",
    "root_mean_squared_error",
    "spearmanr",
    "top_k_accuracy",
    "wasserstein_distance",
]


from arkas.metric.classification.accuracy import accuracy
from arkas.metric.classification.ap import (
    average_precision,
    binary_average_precision,
    multiclass_average_precision,
    multilabel_average_precision,
)
from arkas.metric.classification.balanced_accuracy import balanced_accuracy
from arkas.metric.classification.confmat import (
    binary_confusion_matrix,
    confusion_matrix,
    multiclass_confusion_matrix,
    multilabel_confusion_matrix,
)
from arkas.metric.classification.fbeta import (
    binary_fbeta_score,
    fbeta_score,
    multiclass_fbeta_score,
    multilabel_fbeta_score,
)
from arkas.metric.classification.jaccard import (
    binary_jaccard,
    jaccard,
    multiclass_jaccard,
    multilabel_jaccard,
)
from arkas.metric.classification.ndcg import ndcg
from arkas.metric.classification.precision import (
    binary_precision,
    multiclass_precision,
    multilabel_precision,
    precision,
)
from arkas.metric.classification.recall import (
    binary_recall,
    multiclass_recall,
    multilabel_recall,
    recall,
)
from arkas.metric.classification.roc_auc import (
    binary_roc_auc,
    multiclass_roc_auc,
    multilabel_roc_auc,
    roc_auc,
)
from arkas.metric.classification.topk_accuracy import (
    binary_top_k_accuracy,
    multiclass_top_k_accuracy,
    top_k_accuracy,
)
from arkas.metric.correlation.pearson import pearsonr
from arkas.metric.correlation.spearman import spearmanr
from arkas.metric.distribution.energy import energy_distance
from arkas.metric.distribution.jensen_shannon import jensen_shannon_divergence
from arkas.metric.distribution.kl import kl_div
from arkas.metric.distribution.wasserstein import wasserstein_distance
from arkas.metric.regression.abs_error import mean_absolute_error, median_absolute_error
from arkas.metric.regression.mape import mean_absolute_percentage_error
from arkas.metric.regression.mse import mean_squared_error
from arkas.metric.regression.msle import mean_squared_log_error
from arkas.metric.regression.r2 import r2_score
from arkas.metric.regression.rmse import root_mean_squared_error
from arkas.metric.regression.tweedie_deviance import mean_tweedie_deviance
from arkas.metric.regression.universal import regression_errors
